#include <config.h>

#include "cf_defs.h"
#include "canonicalform.h"
#include "fac_util.h"
#include "cf_algorithm.h"
#include "cf_reval.h"
#include "cf_random.h"
#include "cf_primes.h"
#include "fac_distrib.h"
#include "ftmpl_functions.h"
#include "cf_reval.h"


static bool fin_findeval( const CanonicalForm & F, const CanonicalForm & G, CanonicalForm & Fb, CanonicalForm & Gb, CanonicalForm & Db, REvaluation & b, int delta, int degF, int degG, int bound, int & counter );

CanonicalForm fin_ezgcd( const CanonicalForm & FF, const CanonicalForm & GG )
{
  CanonicalForm F, G, f, g, d, Fb, Gb, Db, Fbt, Gbt, Dbt, B0, B, D0, lcF, lcG, lcD;
  CFArray DD( 1, 2 ), lcDD( 1, 2 );
  int degF, degG, delta, count, maxeval;
  count = 0; // number of eval. used
  maxeval = getCharacteristic()/2; // bound on the number of eval. to use
  REvaluation b, bt;
  bool gcdfound = false;
  Variable x = Variable(1);
  modpk p = modpk( getCharacteristic(), 1 ); // p = p^k with k=1
  f = content( FF, x ); g = content( GG, x ); d = gcd( f, g );
  F = FF / f; G = GG / g;
  if( F.isUnivariate() && G.isUnivariate() )
  {
    if( F.mvar() == G.mvar() )
      d *= gcd( F, G );
    return d;
  }
  lcF = LC( F, x ); lcG = LC( G, x );
  lcD = gcd( lcF, lcG );
  delta = 0;
  degF = degree( F, x ); degG = degree( G, x );
  b = REvaluation( 2, tmax(F.level(), G.level()), FFRandom() );
  while( !gcdfound )
  {
    if( !fin_findeval( F, G, Fb, Gb, Db, b, delta, degF, degG, maxeval, count ))
    { // too many eval. used --> try another method
      Off( SW_USE_EZGCD_P );
      d *= gcd( F, G );
      On( SW_USE_EZGCD_P );
      return d;
    }
    delta = degree( Db );
    if( delta == 0 )
      return d;
    while( true )
    {
      bt = b;
      if( !fin_findeval( F, G, Fbt, Gbt, Dbt, bt, delta+1, degF, degG, maxeval, count ))
      { // too many eval. used --> try another method
        Off( SW_USE_EZGCD_P );
        d *= gcd( F, G );
        On( SW_USE_EZGCD_P );
        return d;
      }
      int dd = degree( Dbt );
      if( dd == 0 )
        return d;
      if( dd == delta )
        break;
      if( dd < delta )
      {
        delta = dd;
        b = bt;
        Db = Dbt; Fb = Fbt; Gb = Gbt;
      }
    }
    if( degF <= degG && delta == degF && fdivides( F, G ) )
      return d*F;
    if( degG < degF && delta == degG && fdivides( G, F ) )
      return d*G;
    if( delta != degF && delta != degG )
    {
      bool B_is_F;
      CanonicalForm xxx;
      DD[1] = Fb / Db;
      xxx = gcd( DD[1], Db );
      if( xxx.inCoeffDomain() )
      {
        B = F;
        DD[2] = Db;
        lcDD[1] = lcF;
        lcDD[2] = lcF;
        B *= lcF;
        B_is_F = true;
      }
      else
      {
        DD[1] = Gb / Db;
        xxx = gcd( DD[1], Db );
        if( xxx.inCoeffDomain() )
        {
          B = G;
          DD[2] = Db;
          lcDD[1] = lcG;
          lcDD[2] = lcG;
          B *= lcG;
          B_is_F = false;
        }
        else // special case handling
        {
          Off( SW_USE_EZGCD_P );
          d *= gcd( F, G ); // try another method
          On( SW_USE_EZGCD_P );
          return d;
        }
      }
      DD[2] = DD[2] * ( b( lcDD[2] ) / lc( DD[2] ) );
      DD[1] = DD[1] * ( b( lcDD[1] ) / lc( DD[1] ) );
      
      gcdfound = Hensel( B, DD, lcDD, b, p, x );

      if( gcdfound )
      {
        CanonicalForm cand = DD[2] / content( DD[2], Variable(1) );
        if( B_is_F )
          gcdfound = fdivides( cand, G );
        else
          gcdfound = fdivides( cand, F );
      }
    }
    delta++;
  }
  return d * DD[2] / content( DD[2],Variable(1) );
}


static bool fin_findeval( const CanonicalForm & F, const CanonicalForm & G, CanonicalForm & Fb, CanonicalForm & Gb, CanonicalForm & Db, REvaluation & b, int delta, int degF, int degG, int maxeval, int & count )
{
  if( delta != 0 )
  {
    b.nextpoint();
    if( count++ > maxeval )
      return false; // too many eval. used
  }
  while( true )
  {
    Fb = b( F );
    if( degree( Fb ) == degF )
    {
      Gb = b( G );
      if( degree( Gb ) == degG )
      {
        Db = gcd( Fb, Gb );
        if( delta > 0 )
        {
          if( degree( Db ) < delta )
            break;
        }
        else break;
      }
    }
    b.nextpoint();
    if( count++ > maxeval ) // too many eval. used
      return false;
  }
  return true;
}
